from __future__ import unicode_literals

from collections import OrderedDict

from django.template import loader

from rest_framework.compat import template_render
from rest_framework.exceptions import NotFound
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.utils.urls import remove_query_param, replace_query_param
from rest_framework.pagination import _divide_with_ceil, _get_displayed_page_numbers, _get_page_links, _positive_int


class BaseDataFramePagination(object):
    display_page_controls = False

    def paginate_dataframe(self, dataframe, request, view=None):  # pragma: no cover
        raise NotImplementedError('paginate_dataframe() must be implemented.')

    def get_paginated_response(self, data):  # pragma: no cover
        raise NotImplementedError('get_paginated_response() must be implemented.')

    def to_html(self):  # pragma: no cover
        raise NotImplementedError('to_html() must be implemented to display page controls.')

    def get_results(self, data):
        return data['results']

    def get_fields(self, view):
        return []


class LimitOffsetPagination(BaseDataFramePagination):
    """
    A limit/offset based style. For example:

    http://api.example.org/accounts/?limit=100
    http://api.example.org/accounts/?offset=400&limit=100
    """
    default_limit = api_settings.PAGE_SIZE
    limit_query_param = 'limit'
    offset_query_param = 'offset'
    max_limit = None
    template = 'rest_framework/pagination/numbers.html'

    def paginate_dataframe(self, dataframe, request, view=None):
        self.limit = self.get_limit(request)
        if self.limit is None:
            return None

        self.offset = self.get_offset(request)
        self.count = len(dataframe)
        self.request = request
        if self.count > self.limit and self.template is not None:
            self.display_page_controls = True

        if self.count == 0 or self.offset > self.count:
            return []
        return dataframe[self.offset:self.offset + self.limit]

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('count', self.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))

    def get_limit(self, request):
        if self.limit_query_param:
            try:
                return _positive_int(
                    request.query_params[self.limit_query_param],
                    strict=True,
                    cutoff=self.max_limit
                )
            except (KeyError, ValueError):
                pass

        return self.default_limit

    def get_offset(self, request):
        try:
            return _positive_int(
                request.query_params[self.offset_query_param],
            )
        except (KeyError, ValueError):
            return 0

    def get_next_link(self):
        if self.offset + self.limit >= self.count:
            return None

        url = self.request.build_absolute_uri()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        offset = self.offset + self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_previous_link(self):
        if self.offset <= 0:
            return None

        url = self.request.build_absolute_uri()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        if self.offset - self.limit <= 0:
            return remove_query_param(url, self.offset_query_param)

        offset = self.offset - self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_html_context(self):
        base_url = self.request.build_absolute_uri()

        if self.limit:
            current = _divide_with_ceil(self.offset, self.limit) + 1

            # The number of pages is a little bit fiddly.
            # We need to sum both the number of pages from current offset to end
            # plus the number of pages up to the current offset.
            # When offset is not strictly divisible by the limit then we may
            # end up introducing an extra page as an artifact.
            final = (
                _divide_with_ceil(self.count - self.offset, self.limit) +
                _divide_with_ceil(self.offset, self.limit)
            )

            if final < 1:
                final = 1
        else:
            current = 1
            final = 1

        if current > final:
            current = final

        def page_number_to_url(page_number):
            if page_number == 1:
                return remove_query_param(base_url, self.offset_query_param)
            else:
                offset = self.offset + ((page_number - current) * self.limit)
                return replace_query_param(base_url, self.offset_query_param, offset)

        page_numbers = _get_displayed_page_numbers(current, final)
        page_links = _get_page_links(page_numbers, current, page_number_to_url)

        return {
            'previous_url': self.get_previous_link(),
            'next_url': self.get_next_link(),
            'page_links': page_links
        }

    def to_html(self):
        template = loader.get_template(self.template)
        context = self.get_html_context()
        return template_render(template, context)

    def get_fields(self, view):
        return [self.limit_query_param, self.offset_query_param]



import pandas as pd

from collections import OrderedDict

from rest_framework.serializers import (CharField, JSONField, ListField, Serializer,
                                        ValidationError, api_settings)


class DataFrameReadOnlyToDictRecordsSerializer(Serializer):
    """
    A read-only Serializer implementation that uses
    :func:`pandas.DataFrame.to_dict <pandas.DataFrame.to_dict>` to convert data to external
    representation with 'records' orientation. This serializer is useful when a list of dictionaries
    is required.
    """
    def to_internal_value(self, data):
        raise NotImplementedError('`to_representation()` must be implemented.')

    def to_representation(self, instance):
        return {'records': instance.to_dict(orient='records')}


class DataFrameListSerializer(Serializer):
    """
    A Serializer implementation that uses
    :func:`pandas.DataFrame.from_dict <pandas.DataFrame.from_dict>` and
    :func:`pandas.DataFrame.to_dict <pandas.DataFrame.to_dict>` to convert data to internal value
    and to external representation with orientation 'columns' and 'list'.
    """
    def to_internal_value(self, data):
        try:
            return pd.DataFrame.from_dict(data, orient='columns')
        except ValueError as e:
            raise ValidationError({api_settings.NON_FIELD_ERRORS_KEY: [str(e)]})

    def to_representation(self, instance):
        return instance.to_dict(orient='list')


class DataFrameIndexSerializer(Serializer):
    """
    A Serializer implementation that uses
    :func:`pandas.DataFrame.from_dict <pandas.DataFrame.from_dict>` and
    :func:`pandas.DataFrame.to_dict <pandas.DataFrame.to_dict>` to convert data to internal value
    and to external representation with orientation 'index'.
    """
    def to_internal_value(self, data):
        try:
            data_frame = pd.DataFrame.from_dict(data, orient='index').rename(index=int)
            return data_frame
        except ValueError as e:
            raise ValidationError({api_settings.NON_FIELD_ERRORS_KEY: [str(e)]})

    def to_representation(self, instance):
        instance = instance.rename(index=str)
        return instance.to_dict(orient='index')


class DataFrameRecordsSerializer(Serializer):
    """
    A Serializer implementation that uses
    :func:`pandas.DataFrame.from_dict <pandas.DataFrame.from_records>` and
    :func:`pandas.DataFrame.to_dict <pandas.DataFrame.to_records>` to convert data to internal
    value and to external representation.
    """
    columns = ListField(child=CharField())
    data = ListField(child=JSONField())

    def to_internal_value(self, data):
        validated_data = super().to_internal_value(data)

        try:
            data_frame = pd.DataFrame.from_records(validated_data['data'], index='index',
                                                   columns=validated_data['columns'])
            return data_frame
        except ValueError as e:
            raise ValidationError({api_settings.NON_FIELD_ERRORS_KEY: [str(e)]})

    def to_representation(self, instance):
        recarray = instance.to_records(index=True)
        return OrderedDict([('columns', recarray.dtype.names), ('data', recarray.tolist())])



"""
Basic building blocks for generic class based views.

We don't bind behaviour to http method handlers yet,
which allows mixin classes to be composed in interesting ways.
"""
from __future__ import unicode_literals

from rest_framework import status
from rest_framework.response import Response
from rest_framework.settings import api_settings


class CreateDataFrameMixin(object):
    """
    Adds a row to the dataframe.
    """
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create(self, serializer):
        dataframe = self.get_dataframe()
        return self.update_dataframe(dataframe.append(serializer.validated_data))

    def get_success_headers(self, data):
        try:
            return {'Location': data[api_settings.URL_FIELD_NAME]}
        except (TypeError, KeyError):
            return {}


class ListDataFrameMixin(object):
    """
    List the contents of a dataframe.
    """
    def list(self, request, *args, **kwargs):
        dataframe = self.filter_dataframe(self.get_dataframe())

        page = self.paginate_dataframe(dataframe)
        if page is not None:
            serializer = self.get_serializer(page)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(dataframe)
        return Response(serializer.data)


class RetrieveDataFrameMixin(object):
    """
    Retrieve a dataframe row.
    """
    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class UpdateDataFrameMixin(object):
    """
    Update a dataframe row.
    """
    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(instance, serializer)
        return Response(serializer.data)

    def perform_update(self, instance, serializer):
        validated_data = serializer.validated_data
        instance.ix[validated_data.index, validated_data.columns] = validated_data[:]
        dataframe = self.get_dataframe()
        dataframe.ix[instance.index] = instance
        return self.update_dataframe(dataframe)

    def partial_update(self, request, *args, **kwargs):
        kwargs['partial'] = True
        return self.update(request, *args, **kwargs)


class DestroyDataFrameMixin(object):
    """
    Destroy a dataframe row.
    """
    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance):
        dataframe = self.get_dataframe()
        return self.update_dataframe(dataframe.drop(instance.index))



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pandas_drf_tools',
    version='0.1.1',
    description='A set of tools to make Pandas easy to use with Django REST Framework projects',
    long_description=long_description,
    url='https://github.com/abarto/pandas-drf-tools',
    author='Agustin Barto',
    author_email='abarto@gmail.com',
    license='BSD',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    keywords='pandas djangorestframework django',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        'pandas>=0.18.1',
        'djangorestframework>=3.4.6'
    ],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage', 'flake8', 'nose']
    }
)



from rest_framework.viewsets import ViewSetMixin

from pandas_drf_tools import generics, mixins


class GenericDataFrameViewSet(ViewSetMixin, generics.GenericDataFrameAPIView):
    """
    The GenericDataFrameViewSet class does not provide any actions by default,
    but does include the base set of generic view behavior, such as
    the `get_object` and `get_dataframe` methods.
    """
    pass


class ReadOnlyDataFrameViewSet(mixins.RetrieveDataFrameMixin,
                               mixins.ListDataFrameMixin,
                               GenericDataFrameViewSet):
    """
    A viewset that provides default `list()` and `retrieve()` actions.
    """
    pass


class DataFrameViewSet(mixins.CreateDataFrameMixin,
                       mixins.RetrieveDataFrameMixin,
                       mixins.UpdateDataFrameMixin,
                       mixins.DestroyDataFrameMixin,
                       mixins.ListDataFrameMixin,
                       GenericDataFrameViewSet):
    """
    A viewset that provides default `create()`, `retrieve()`, `update()`,
    `partial_update()`, `destroy()` and `list()` actions.
    """
    pass



from django.http import Http404

from rest_framework.views import APIView

from pandas_drf_tools import mixins


class GenericDataFrameAPIView(APIView):
    """Base class for all other generic DataFrame views. It is based on GenericAPIView."""
    # You'll need to either set these attributes,
    # or override `get_dataframe()`/`get_serializer_class()`.
    # If you are overriding a view method, it is important that you call
    # `get_dataframe()` instead of accessing the `dataframe` property directly,
    # as `dataframe` will get evaluated only once, and those results are cached
    # for all subsequent requests.
    dataframe = None
    serializer_class = None

    # If you want to use object lookups other than index, set 'lookup_url_kwarg'.
    # For more complex lookup requirements override `get_object()`.
    lookup_url_kwarg = 'index'

    # The style to use for dataframe pagination.
    pagination_class = None

    def get_dataframe(self):
        """
        Get the DataFrame for this view.
        Defaults to using `self.dataframe`.

        This method should always be used rather than accessing `self.dataframe`
        directly, as `self.dataframe` gets evaluated only once, and those results
        are cached for all subsequent requests.

        You may want to override this if you need to provide different
        dataframes depending on the incoming request.
        """
        assert self.dataframe is not None, (
            "'%s' should either include a `dataframe` attribute, "
            "or override the `get_dataframe()` method."
            % self.__class__.__name__
        )

        dataframe = self.dataframe
        return dataframe

    def update_dataframe(self, dataframe):
        """
        Indicates that the dataframe needs to be updated. The default implementation
        just returns the argument. This method has to be ovewritten to make changing
        operations stick.
        """
        return dataframe

    def index_row(self, dataframe):
        """
        Indexes the row based on the request parameters.
        """
        return dataframe.loc[self.kwargs[self.lookup_url_kwarg]].to_frame().T

    def get_object(self):
        """
        Returns the row the view is displaying.

        You may want to override this if you need to provide non-standard
        queryset lookups.  Eg if objects are referenced using multiple
        keyword arguments in the url conf.
        """
        dataframe = self.filter_dataframe(self.get_dataframe())

        assert self.lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            (self.__class__.__name__, self.lookup_url_kwarg)
        )

        try:
            obj = self.index_row(dataframe)
        except (IndexError, KeyError, ValueError):
            raise Http404

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.

        You may want to override this if you need to provide different
        serializations depending on the incoming request.

        (Eg. admins get full serialization, others get basic serialization)
        """
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )

        return self.serializer_class

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }

    def filter_dataframe(self, dataframe):
        """
        Given a dataframe, filter it.
        """
        return dataframe

    @property
    def paginator(self):
        """
        The paginator instance associated with the view, or `None`.
        """
        if not hasattr(self, '_paginator'):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_dataframe(self, dataframe):
        """
        Return a single page of results, or `None` if pagination is disabled.
        """
        if self.paginator is None:
            return None
        return self.paginator.paginate_dataframe(dataframe, self.request, view=self)

    def get_paginated_response(self, data):
        """
        Return a paginated style `Response` object for the given output data.
        """
        assert self.paginator is not None
        return self.paginator.get_paginated_response(data)


# Concrete view classes that provide method handlers
# by composing the mixin classes with the base view.

class CreateAPIView(mixins.CreateDataFrameMixin,
                    GenericDataFrameAPIView):
    """
    Concrete view for creating a model instance.
    """
    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class ListAPIView(mixins.ListDataFrameMixin,
                  GenericDataFrameAPIView):
    """
    Concrete view for listing a queryset.
    """
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class RetrieveAPIView(mixins.RetrieveDataFrameMixin,
                      GenericDataFrameAPIView):
    """
    Concrete view for retrieving a model instance.
    """
    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)


class DestroyAPIView(mixins.DestroyDataFrameMixin,
                     GenericDataFrameAPIView):
    """
    Concrete view for deleting a model instance.
    """
    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class UpdateAPIView(mixins.UpdateDataFrameMixin,
                    GenericDataFrameAPIView):
    """
    Concrete view for updating a model instance.
    """
    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class ListCreateAPIView(mixins.ListDataFrameMixin,
                        mixins.CreateDataFrameMixin,
                        GenericDataFrameAPIView):
    """
    Concrete view for listing a queryset or creating a model instance.
    """
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class RetrieveUpdateAPIView(mixins.RetrieveDataFrameMixin,
                            mixins.UpdateDataFrameMixin,
                            GenericDataFrameAPIView):
    """
    Concrete view for retrieving, updating a model instance.
    """
    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class RetrieveDestroyAPIView(mixins.RetrieveDataFrameMixin,
                             mixins.DestroyDataFrameMixin,
                             GenericDataFrameAPIView):
    """
    Concrete view for retrieving or deleting a model instance.
    """
    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class RetrieveUpdateDestroyAPIView(mixins.RetrieveDataFrameMixin,
                                   mixins.UpdateDataFrameMixin,
                                   mixins.DestroyDataFrameMixin,
                                   GenericDataFrameAPIView):
    """
    Concrete view for retrieving, updating or deleting a model instance.
    """
    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)
